use anyhow::Result;
use clap::Parser;
use http_body_util::Full;
use hyper::body::{Bytes, Incoming};
use hyper::client::conn::http1::SendRequest;
use hyper::{Request, Response};
use rand::prelude::Distribution;
use rand::{distributions::Uniform, Rng};
use serde::Serialize;
use std::net::SocketAddr;
use std::sync::Arc;
use tokio::io::{AsyncWriteExt, BufReader};
use tokio::net::tcp::{ReadHalf, WriteHalf};
use tokio::net::TcpStream;
use tokio::io::{AsyncBufReadExt, AsyncWriteExt, BufReader};

#[derive(Parser)]
struct CliArgs {
    /// Management API address
    #[arg(long)]
    mgmt: SocketAddr,
    /// IRC server address
    #[arg(long)]
    irc: SocketAddr,
}

#[tokio::main]
async fn main() -> Result<()> {
    tracing_subscriber::fmt::init();
    let args = CliArgs::parse();
    let mut gen = rand::thread_rng();

    tracing::info!("Starting");
    let stream = TcpStream::connect(args.mgmt).await?;
    let (mut request_sender, connection) = hyper::client::conn::http1::handshake(stream).await?;
    tracing::info!("Mgmt connected");

    let fiber = tokio::spawn(async move {
        if let Err(e) = connection.await {
            eprintln!("Error in connection: {}", e);
        }
    });

    let name = generate_player_name(&mut gen);
    let pass = generate_password(&mut gen);
    {
        let v = mgmt_api::CreatePlayerRequest { name: &*name };
        let response = http_send(&mut request_sender, mgmt_api::paths::CREATE_PLAYER, &v).await?;
        assert!(response.status().is_success());
        tracing::info!("Player {} created", name);
    }
    {
        let v = mgmt_api::ChangePasswordRequest {
            player_name: &*name,
            password: &*pass,
        };
        let response = http_send(&mut request_sender, mgmt_api::paths::SET_PASSWORD, &v).await?;
        assert!(response.status().is_success());
        tracing::info!("Password set for {}", name);
    }
    drop(request_sender);
    fiber.await?;

    let mut stream = TcpStream::connect(args.irc).await?;
    tracing::info!("IRC connected");
    let (reader, writer) = stream.split();
    let reader = BufReader::new(reader);
    let test_scope = TestScope { reader, writer, buffer: vec![] };

    test_scope.send(&format!("PASS {pass}")).await?;
    test_scope.send(&format!("NICK {name}")).await?;
    test_scope.send(&format!("USER UserName 0 * :Real Name")).await?;

    stream.shutdown().await?;
    Ok(())
}

struct TestScope<'a> {
    reader: BufReader<ReadHalf<'a>>,
    writer: WriteHalf<'a>,
    buffer: Vec<u8>,
}
impl<'a> TestScope<'a> {
    async fn send(&mut self, str: &(impl AsRef<str> + ?Sized)) -> Result<()> {
        self.writer.write_all(str.as_ref().as_bytes()).await?;
        self.writer.write_all(b"\r\n").await?;
        self.writer.flush().await?;
        Ok(())
    }

    async fn assert(&mut self, f: impl FnOnce(&str) -> Result<()>) -> Result<()> {
        let len = self.reader.read_until(b'\n', &mut self.buffer).await?;
        let res = f(std::str::from_utf8(&self.buffer[0..len])?);
        self.buffer.clear();
        res
    }
}

async fn http_send(
    client: &mut SendRequest<Full<Bytes>>,
    path: &str,
    payload: &impl Serialize,
) -> Result<Response<Incoming>> {
    let mut buffer = vec![];
    serde_json::to_writer(&mut buffer, payload).expect("unexpected fail when writing to vec");
    let body = Full::new(Bytes::from(buffer));
    let request = Request::builder().method("POST").uri(path).body(body)?;
    let response = client.send_request(request).await?;
    Ok(response)
}

fn generate_player_name(gen: &mut impl Rng) -> Arc<str> {
    let len = Uniform::from(7..14).sample(gen);
    gen.sample_iter(Uniform::from('a'..'z'))
        .take(len)
        .collect::<String>()
        .into()
}

fn generate_password(gen: &mut impl Rng) -> Arc<str> {
    let len = Uniform::from(8..32).sample(gen);
    gen.sample_iter(Uniform::from('0'..'z'))
        .take(len)
        .collect::<String>()
        .into()
}
